@@ -33,14 +33,18 @@
 #include <coreplugin/editormanager/documentmodel.h>
 #include <coreplugin/icore.h>
 #include <coreplugin/idocument.h>
-#include <coreplugin/variablechooser.h>
+
 #include <projectexplorer/project.h>
 #include <projectexplorer/session.h>
+
 #include <utils/algorithm.h>
+#include <utils/utilsicons.h>
 #include <utils/delegates.h>
 #include <utils/fancylineedit.h>
 #include <utils/mimetypes/mimedatabase.h>
 #include <utils/jsontreeitem.h>
+#include <utils/stringutils.h>
+#include <utils/variablechooser.h>
 
 #include <QBoxLayout>
 #include <QComboBox>
@@ -51,24 +55,32 @@
 #include <QDir>
 #include <QFileInfo>
 #include <QHeaderView>
+#include <QJsonDocument>
 #include <QLabel>
 #include <QListView>
+#include <QMenu>
+#include <QMimeData>
 #include <QPushButton>
 #include <QSettings>
 #include <QSortFilterProxyModel>
 #include <QStringListModel>
+#include <QToolButton>
 #include <QTreeView>
 
+constexpr char typeIdKey[] = "typeId";
 constexpr char nameKey[] = "name";
 constexpr char idKey[] = "id";
 constexpr char enabledKey[] = "enabled";
 constexpr char startupBehaviorKey[] = "startupBehavior";
 constexpr char mimeTypeKey[] = "mimeType";
 constexpr char filePatternKey[] = "filePattern";
+constexpr char initializationOptionsKey[] = "initializationOptions";
 constexpr char executableKey[] = "executable";
 constexpr char argumentsKey[] = "arguments";
 constexpr char settingsGroupKey[] = "LanguageClient";
 constexpr char clientsKey[] = "clients";
+constexpr char typedClientsKey[] = "typedClients";
+constexpr char mimeType[] = "application/language.client.setting";
 
 namespace LanguageClient {
 
@@ -85,14 +97,25 @@ public:
     bool insertRows(int row, int count = 1, const QModelIndex &parent = QModelIndex()) final;
     bool setData(const QModelIndex &index, const QVariant &value, int role) final;
     Qt::ItemFlags flags(const QModelIndex &index) const final;
+    Qt::DropActions supportedDropActions() const override { return Qt::MoveAction; }
+    QStringList mimeTypes() const override { return {mimeType}; }
+    QMimeData *mimeData(const QModelIndexList &indexes) const override;
+    bool dropMimeData(const QMimeData *data,
+                      Qt::DropAction action,
+                      int row,
+                      int column,
+                      const QModelIndex &parent) override;
 
     void reset(const QList<BaseSettings *> &settings);
     QList<BaseSettings *> settings() const { return m_settings; }
+    int insertSettings(BaseSettings *settings);
+    void enableSetting(const QString &id);
     QList<BaseSettings *> removed() const { return m_removed; }
     BaseSettings *settingForIndex(const QModelIndex &index) const;
     QModelIndex indexForSetting(BaseSettings *setting) const;
 
 private:
+    static constexpr int idRole = Qt::UserRole + 1;
     QList<BaseSettings *> m_settings; // owned
     QList<BaseSettings *> m_removed;
 };
@@ -114,7 +137,7 @@ private:
         QWidget *widget = nullptr;
     } m_currentSettings;
 
-    void addItem();
+    void addItem(const Utils::Id &clientTypeId);
     void deleteItem();
 };
 
@@ -133,12 +156,22 @@ public:
     void finish() override;
 
     QList<BaseSettings *> settings() const;
+    QList<BaseSettings *> changedSettings() const;
+    void addSettings(BaseSettings *settings);
+    void enableSettings(const QString &id);
 
 private:
     LanguageClientSettingsModel m_model;
+    QSet<QString> m_changedSettings;
     QPointer<LanguageClientSettingsPageWidget> m_widget;
 };
 
+QMap<Utils::Id, ClientType> &clientTypes()
+{
+    static QMap<Utils::Id, ClientType> types;
+    return types;
+}
+
 LanguageClientSettingsPageWidget::LanguageClientSettingsPageWidget(LanguageClientSettingsModel &settings)
     : m_settings(settings)
     , m_view(new QTreeView())
@@ -149,11 +182,22 @@ LanguageClientSettingsPageWidget::LanguageClientSettingsPageWidget(LanguageClien
     m_view->setHeaderHidden(true);
     m_view->setSelectionMode(QAbstractItemView::SingleSelection);
     m_view->setSelectionBehavior(QAbstractItemView::SelectItems);
+    m_view->setDragEnabled(true);
+    m_view->viewport()->setAcceptDrops(true);
+    m_view->setDropIndicatorShown(true);
+    m_view->setDragDropMode(QAbstractItemView::InternalMove);
     connect(m_view->selectionModel(), &QItemSelectionModel::currentChanged,
             this, &LanguageClientSettingsPageWidget::currentChanged);
     auto buttonLayout = new QVBoxLayout();
     auto addButton = new QPushButton(LanguageClientSettingsPage::tr("&Add"));
-    connect(addButton, &QPushButton::pressed, this, &LanguageClientSettingsPageWidget::addItem);
+    auto addMenu = new QMenu;
+    addMenu->clear();
+    for (const ClientType &type : clientTypes()) {
+        auto action = new QAction(LanguageClientSettingsPage::tr("New %1").arg(type.name));
+        connect(action, &QAction::triggered, this, [this, id = type.id]() { addItem(id); });
+        addMenu->addAction(action);
+    }
+    addButton->setMenu(addMenu);
     auto deleteButton = new QPushButton(LanguageClientSettingsPage::tr("&Delete"));
     connect(deleteButton, &QPushButton::pressed, this, &LanguageClientSettingsPageWidget::deleteItem);
     mainLayout->addLayout(layout);
@@ -205,23 +249,36 @@ void LanguageClientSettingsPageWidget::applyCurrentSettings()
     if (!m_currentSettings.setting)
         return;
 
-    m_currentSettings.setting->applyFromSettingsWidget(m_currentSettings.widget);
-    auto index = m_settings.indexForSetting(m_currentSettings.setting);
-    emit m_settings.dataChanged(index, index);
+    if (m_currentSettings.setting->applyFromSettingsWidget(m_currentSettings.widget)) {
+        auto index = m_settings.indexForSetting(m_currentSettings.setting);
+        emit m_settings.dataChanged(index, index);
+    }
 }
 
-void LanguageClientSettingsPageWidget::addItem()
+BaseSettings *generateSettings(const Utils::Id &clientTypeId)
 {
-    const int row = m_settings.rowCount();
-    m_settings.insertRows(row);
-    m_view->setCurrentIndex(m_settings.index(row));
+    if (auto generator = clientTypes().value(clientTypeId).generator) {
+        auto settings = generator();
+        settings->m_settingsTypeId = clientTypeId;
+        return settings;
+    }
+    return nullptr;
+}
+
+void LanguageClientSettingsPageWidget::addItem(const Utils::Id &clientTypeId)
+{
+    auto newSettings = generateSettings(clientTypeId);
+    QTC_ASSERT(newSettings, return);
+    m_view->setCurrentIndex(m_settings.index(m_settings.insertSettings(newSettings)));
 }
 
 void LanguageClientSettingsPageWidget::deleteItem()
 {
     auto index = m_view->currentIndex();
-    if (index.isValid())
-        m_settings.removeRows(index.row());
+    if (!index.isValid())
+        return;
+
+    m_settings.removeRows(index.row());
 }
 
 LanguageClientSettingsPage::LanguageClientSettingsPage()
@@ -231,8 +288,11 @@ LanguageClientSettingsPage::LanguageClientSettingsPage()
     setCategory(Constants::LANGUAGECLIENT_SETTINGS_CATEGORY);
     setDisplayCategory(QCoreApplication::translate("LanguageClient",
                                                    Constants::LANGUAGECLIENT_SETTINGS_TR));
-    setCategoryIcon(Utils::Icon({{":/languageclient/images/settingscategory_languageclient.png",
-                    Utils::Theme::PanelTextColorDark}}, Utils::Icon::Tint));
+    setCategoryIconPath(":/languageclient/images/settingscategory_languageclient.png");
+    connect(&m_model, &LanguageClientSettingsModel::dataChanged, [this](const QModelIndex &index) {
+        if (BaseSettings *setting = m_model.settingForIndex(index))
+            m_changedSettings << setting->m_id;
+    });
 }
 
 LanguageClientSettingsPage::~LanguageClientSettingsPage()
@@ -278,6 +338,7 @@ void LanguageClientSettingsPage::apply()
 void LanguageClientSettingsPage::finish()
 {
     m_model.reset(LanguageClientManager::currentSettings());
+    m_changedSettings.clear();
 }
 
 QList<BaseSettings *> LanguageClientSettingsPage::settings() const
@@ -285,6 +346,27 @@ QList<BaseSettings *> LanguageClientSettingsPage::settings() const
     return m_model.settings();
 }
 
+QList<BaseSettings *> LanguageClientSettingsPage::changedSettings() const
+{
+    QList<BaseSettings *> result;
+    const QList<BaseSettings *> &all = settings();
+    for (BaseSettings *setting : all) {
+        if (m_changedSettings.contains(setting->m_id))
+            result << setting;
+    }
+    return result;
+}
+
+void LanguageClientSettingsPage::addSettings(BaseSettings *settings)
+{
+    m_model.insertSettings(settings);
+}
+
+void LanguageClientSettingsPage::enableSettings(const QString &id)
+{
+    m_model.enableSetting(id);
+}
+
 LanguageClientSettingsModel::~LanguageClientSettingsModel()
 {
     qDeleteAll(m_settings);
@@ -299,6 +381,8 @@ QVariant LanguageClientSettingsModel::data(const QModelIndex &index, int role) c
         return Utils::globalMacroExpander()->expand(setting->m_name);
     else if (role == Qt::CheckStateRole)
         return setting->m_enabled ? Qt::Checked : Qt::Unchecked;
+    else if (role == idRole)
+        return setting->m_id;
     return QVariant();
 }
 
@@ -338,9 +422,55 @@ bool LanguageClientSettingsModel::setData(const QModelIndex &index, const QVaria
     return true;
 }
 
-Qt::ItemFlags LanguageClientSettingsModel::flags(const QModelIndex &/*index*/) const
+Qt::ItemFlags LanguageClientSettingsModel::flags(const QModelIndex &index) const
 {
-    return Qt::ItemIsSelectable | Qt::ItemIsEnabled | Qt::ItemIsUserCheckable;
+    const Qt::ItemFlags dragndropFlags = index.isValid() ? Qt::ItemIsDragEnabled
+                                                         : Qt::ItemIsDropEnabled;
+    return Qt::ItemIsSelectable | Qt::ItemIsEnabled | Qt::ItemIsUserCheckable | dragndropFlags;
+}
+
+QMimeData *LanguageClientSettingsModel::mimeData(const QModelIndexList &indexes) const
+{
+    QTC_ASSERT(indexes.count() == 1, return nullptr);
+
+    QMimeData *mimeData = new QMimeData;
+    QByteArray encodedData;
+
+    QDataStream stream(&encodedData, QIODevice::WriteOnly);
+
+    for (const QModelIndex &index : indexes) {
+        if (index.isValid())
+            stream << data(index, idRole).toString();
+    }
+
+    mimeData->setData(mimeType, indexes.first().data(idRole).toString().toUtf8());
+    return mimeData;
+}
+
+bool LanguageClientSettingsModel::dropMimeData(
+    const QMimeData *data, Qt::DropAction action, int row, int column, const QModelIndex &parent)
+{
+    if (!canDropMimeData(data, action, row, column, parent))
+        return false;
+
+    if (action == Qt::IgnoreAction)
+        return true;
+
+    const QString id = QString::fromUtf8(data->data(mimeType));
+    auto setting = Utils::findOrDefault(m_settings, [id](const BaseSettings *setting) {
+        return setting->m_id == id;
+    });
+    if (!setting)
+        return false;
+
+    if (row == -1)
+        row = parent.isValid() ? parent.row() : rowCount(QModelIndex());
+
+    beginInsertRows(parent, row, row);
+    m_settings.insert(row, setting->copy());
+    endInsertRows();
+
+    return true;
 }
 
 void LanguageClientSettingsModel::reset(const QList<BaseSettings *> &settings)
@@ -353,6 +483,26 @@ void LanguageClientSettingsModel::reset(const QList<BaseSettings *> &settings)
     endResetModel();
 }
 
+int LanguageClientSettingsModel::insertSettings(BaseSettings *settings)
+{
+    int row = rowCount();
+    beginInsertRows(QModelIndex(), row, row);
+    m_settings.insert(row, settings);
+    endInsertRows();
+    return row;
+}
+
+void LanguageClientSettingsModel::enableSetting(const QString &id)
+{
+    BaseSettings *setting = Utils::findOrDefault(m_settings, Utils::equal(&BaseSettings::m_id, id));
+    if (!setting)
+        return;
+    setting->m_enabled = true;
+    const QModelIndex &index = indexForSetting(setting);
+    if (index.isValid())
+        emit dataChanged(index, index, {Qt::CheckStateRole});
+}
+
 BaseSettings *LanguageClientSettingsModel::settingForIndex(const QModelIndex &index) const
 {
     if (!index.isValid() || index.row() >= m_settings.size())
@@ -366,13 +516,34 @@ QModelIndex LanguageClientSettingsModel::indexForSetting(BaseSettings *setting)
     return index < 0 ? QModelIndex() : createIndex(index, 0, setting);
 }
 
-void BaseSettings::applyFromSettingsWidget(QWidget *widget)
+QJsonObject BaseSettings::initializationOptions() const
 {
+    return QJsonDocument::fromJson(Utils::globalMacroExpander()->
+                                   expand(m_initializationOptions).toUtf8()).object();
+}
+
+bool BaseSettings::applyFromSettingsWidget(QWidget *widget)
+{
+    bool changed = false;
     if (auto settingsWidget = qobject_cast<BaseSettingsWidget *>(widget)) {
-        m_name = settingsWidget->name();
-        m_languageFilter = settingsWidget->filter();
-        m_startBehavior = settingsWidget->startupBehavior();
+        if (m_name != settingsWidget->name()) {
+            m_name = settingsWidget->name();
+            changed = true;
+        }
+        if (m_languageFilter != settingsWidget->filter()) {
+            m_languageFilter = settingsWidget->filter();
+            changed = true;
+        }
+        if (m_startBehavior != settingsWidget->startupBehavior()) {
+            m_startBehavior = settingsWidget->startupBehavior();
+            changed = true;
+        }
+        if (m_initializationOptions != settingsWidget->initializationOptions()) {
+            m_initializationOptions = settingsWidget->initializationOptions();
+            changed = true;
+        }
     }
+    return changed;
 }
 
 QWidget *BaseSettings::createSettingsWidget(QWidget *parent) const
@@ -380,18 +551,6 @@ QWidget *BaseSettings::createSettingsWidget(QWidget *parent) const
     return new BaseSettingsWidget(this, parent);
 }
 
-bool BaseSettings::needsRestart() const
-{
-    const QVector<Client *> clients = LanguageClientManager::clientForSetting(this);
-    if (clients.isEmpty())
-        return m_enabled;
-    if (!m_enabled)
-        return true;
-    return Utils::anyOf(clients, [this](const Client *client) {
-        return client->needsRestart(this);
-    });
-}
-
 bool BaseSettings::isValid() const
 {
     return !m_name.isEmpty();
@@ -403,21 +562,30 @@ Client *BaseSettings::createClient()
         return nullptr;
     BaseClientInterface *interface = createInterface();
     QTC_ASSERT(interface, return nullptr);
-    auto *client = new Client(interface);
+    auto *client = createClient(interface);
     client->setName(Utils::globalMacroExpander()->expand(m_name));
     client->setSupportedLanguage(m_languageFilter);
+    client->setInitializationOptions(initializationOptions());
+    client->setActivateDocumentAutomatically(true);
     return client;
 }
 
+Client *BaseSettings::createClient(BaseClientInterface *interface) const
+{
+    return new Client(interface);
+}
+
 QVariantMap BaseSettings::toMap() const
 {
     QVariantMap map;
+    map.insert(typeIdKey, m_settingsTypeId.toSetting());
     map.insert(nameKey, m_name);
     map.insert(idKey, m_id);
     map.insert(enabledKey, m_enabled);
     map.insert(startupBehaviorKey, m_startBehavior);
     map.insert(mimeTypeKey, m_languageFilter.mimeTypes);
     map.insert(filePatternKey, m_languageFilter.filePattern);
+    map.insert(initializationOptionsKey, m_initializationOptions);
     return map;
 }
 
@@ -430,6 +598,8 @@ void BaseSettings::fromMap(const QVariantMap &map)
         map.value(startupBehaviorKey, BaseSettings::RequiresFile).toInt());
     m_languageFilter.mimeTypes = map[mimeTypeKey].toStringList();
     m_languageFilter.filePattern = map[filePatternKey].toStringList();
+    m_languageFilter.filePattern.removeAll(QString()); // remove empty entries
+    m_initializationOptions = map[initializationOptionsKey].toString();
 }
 
 static LanguageClientSettingsPage &settingsPage()
@@ -446,39 +616,84 @@ void LanguageClientSettings::init()
 QList<BaseSettings *> LanguageClientSettings::fromSettings(QSettings *settingsIn)
 {
     settingsIn->beginGroup(settingsGroupKey);
-    auto variants = settingsIn->value(clientsKey).toList();
-    auto settings = Utils::transform(variants, [](const QVariant& var){
-        BaseSettings *settings = new StdIOSettings();
-        settings->fromMap(var.toMap());
-        return settings;
-    });
+    QList<BaseSettings *> result;
+
+    for (auto varList :
+         {settingsIn->value(clientsKey).toList(), settingsIn->value(typedClientsKey).toList()}) {
+        for (const QVariant &var : varList) {
+            const QMap<QString, QVariant> &map = var.toMap();
+            Utils::Id typeId = Utils::Id::fromSetting(map.value(typeIdKey));
+            if (!typeId.isValid())
+                typeId = Constants::LANGUAGECLIENT_STDIO_SETTINGS_ID;
+            if (BaseSettings *settings = generateSettings(typeId)) {
+                settings->fromMap(map);
+                result << settings;
+            }
+        }
+    }
+
     settingsIn->endGroup();
-    return settings;
+    return result;
 }
 
-QList<BaseSettings *> LanguageClientSettings::currentPageSettings()
+QList<BaseSettings *> LanguageClientSettings::pageSettings()
 {
     return settingsPage().settings();
 }
 
+QList<BaseSettings *> LanguageClientSettings::changedSettings()
+{
+    return settingsPage().changedSettings();
+}
+
+void LanguageClientSettings::registerClientType(const ClientType &type)
+{
+    QTC_ASSERT(!clientTypes().contains(type.id), return);
+    clientTypes()[type.id] = type;
+}
+
+void LanguageClientSettings::addSettings(BaseSettings *settings)
+{
+    settingsPage().addSettings(settings);
+}
+
+void LanguageClientSettings::enableSettings(const QString &id)
+{
+    settingsPage().enableSettings(id);
+}
+
 void LanguageClientSettings::toSettings(QSettings *settings,
                                         const QList<BaseSettings *> &languageClientSettings)
 {
     settings->beginGroup(settingsGroupKey);
-    settings->setValue(clientsKey, Utils::transform(languageClientSettings,
-                                                    [](const BaseSettings *setting){
-        return QVariant(setting->toMap());
-    }));
+    auto transform = [](const QList<BaseSettings *> &settings) {
+        return Utils::transform(settings, [](const BaseSettings *setting) {
+            return QVariant(setting->toMap());
+        });
+    };
+    auto isStdioSetting = Utils::equal(&BaseSettings::m_settingsTypeId,
+                                       Utils::Id(Constants::LANGUAGECLIENT_STDIO_SETTINGS_ID));
+    auto [stdioSettings, typedSettings] = Utils::partition(languageClientSettings, isStdioSetting);
+    settings->setValue(clientsKey, transform(stdioSettings));
+    settings->setValue(typedClientsKey, transform(typedSettings));
     settings->endGroup();
 }
 
-void StdIOSettings::applyFromSettingsWidget(QWidget *widget)
+bool StdIOSettings::applyFromSettingsWidget(QWidget *widget)
 {
+    bool changed = false;
     if (auto settingsWidget = qobject_cast<StdIOSettingsWidget *>(widget)) {
-        BaseSettings::applyFromSettingsWidget(settingsWidget);
-        m_executable = settingsWidget->executable();
-        m_arguments = settingsWidget->arguments();
+        changed = BaseSettings::applyFromSettingsWidget(settingsWidget);
+        if (m_executable != settingsWidget->executable()) {
+            m_executable = settingsWidget->executable();
+            changed = true;
+        }
+        if (m_arguments != settingsWidget->arguments()) {
+            m_arguments = settingsWidget->arguments();
+            changed = true;
+        }
     }
+    return changed;
 }
 
 QWidget *StdIOSettings::createSettingsWidget(QWidget *parent) const
@@ -486,19 +701,6 @@ QWidget *StdIOSettings::createSettingsWidget(QWidget *parent) const
     return new StdIOSettingsWidget(this, parent);
 }
 
-bool StdIOSettings::needsRestart() const
-{
-    if (BaseSettings::needsRestart())
-        return true;
-    return Utils::anyOf(LanguageClientManager::clientForSetting(this),
-                        [this](QPointer<Client> client) {
-                            if (auto stdIOInterface = qobject_cast<const StdIOClientInterface *>(
-                                    client->clientInterface()))
-                                return stdIOInterface->needsRestart(this);
-                            return false;
-                        });
-}
-
 bool StdIOSettings::isValid() const
 {
     return BaseSettings::isValid() && !m_executable.isEmpty();
@@ -524,9 +726,18 @@ QString StdIOSettings::arguments() const
     return Utils::globalMacroExpander()->expand(m_arguments);
 }
 
+Utils::CommandLine StdIOSettings::command() const
+{
+    return Utils::CommandLine(Utils::FilePath::fromUserInput(m_executable),
+                              arguments(),
+                              Utils::CommandLine::Raw);
+}
+
 BaseClientInterface *StdIOSettings::createInterface() const
 {
-    return new StdIOClientInterface(m_executable, arguments());
+    auto interface = new StdIOClientInterface;
+    interface->setCommandLine(command());
+    return interface;
 }
 
 class JsonTreeItemDelegate : public QStyledItemDelegate
@@ -549,24 +760,6 @@ public:
     }
 };
 
-static QWidget *createCapabilitiesView(const QJsonValue &capabilities)
-{
-    auto root = new Utils::JsonTreeItem("Capabilities", capabilities);
-    if (root->canFetchMore())
-        root->fetchMore();
-
-    auto capabilitiesModel = new Utils::TreeModel<Utils::JsonTreeItem>(root);
-    capabilitiesModel->setHeader({BaseSettingsWidget::tr("Name"),
-                                  BaseSettingsWidget::tr("Value"),
-                                  BaseSettingsWidget::tr("Type")});
-    auto capabilitiesView = new QTreeView();
-    capabilitiesView->setModel(capabilitiesModel);
-    capabilitiesView->setAlternatingRowColors(true);
-    capabilitiesView->header()->setSectionResizeMode(QHeaderView::ResizeToContents);
-    capabilitiesView->setItemDelegate(new JsonTreeItemDelegate);
-    return capabilitiesView;
-}
-
 static QString startupBehaviorString(BaseSettings::StartBehavior behavior)
 {
     switch (behavior) {
@@ -589,13 +782,16 @@ BaseSettingsWidget::BaseSettingsWidget(const BaseSettings *settings, QWidget *pa
     , m_mimeTypes(new QLabel(settings->m_languageFilter.mimeTypes.join(filterSeparator), this))
     , m_filePattern(new QLineEdit(settings->m_languageFilter.filePattern.join(filterSeparator), this))
     , m_startupBehavior(new QComboBox)
+    , m_initializationOptions(new Utils::FancyLineEdit(this))
 {
     int row = 0;
     auto *mainLayout = new QGridLayout;
+
     mainLayout->addWidget(new QLabel(tr("Name:")), row, 0);
     mainLayout->addWidget(m_name, row, 1);
-    auto chooser = new Core::VariableChooser(this);
+    auto chooser = new Utils::VariableChooser(this);
     chooser->addSupportedWidget(m_name);
+
     mainLayout->addWidget(new QLabel(tr("Language:")), ++row, 0);
     auto mimeLayout = new QHBoxLayout;
     mimeLayout->addWidget(m_mimeTypes);
@@ -605,6 +801,7 @@ BaseSettingsWidget::BaseSettingsWidget(const BaseSettings *settings, QWidget *pa
     mainLayout->addLayout(mimeLayout, row, 1);
     m_filePattern->setPlaceholderText(tr("File pattern"));
     mainLayout->addWidget(m_filePattern, ++row, 1);
+
     mainLayout->addWidget(new QLabel(tr("Startup behavior:")), ++row, 0);
     for (int behavior = 0; behavior < BaseSettings::LastSentinel ; ++behavior)
         m_startupBehavior->addItem(startupBehaviorString(BaseSettings::StartBehavior(behavior)));
@@ -615,31 +812,32 @@ BaseSettingsWidget::BaseSettingsWidget(const BaseSettings *settings, QWidget *pa
     connect(addMimeTypeButton, &QPushButton::pressed,
             this, &BaseSettingsWidget::showAddMimeTypeDialog);
 
-    auto createInfoLabel = []() {
-        return new QLabel(tr("Available after server was initialized"));
-    };
+    mainLayout->addWidget(new QLabel(tr("Initialization options:")), ++row, 0);
+    mainLayout->addWidget(m_initializationOptions, row, 1);
+    chooser->addSupportedWidget(m_initializationOptions);
+    m_initializationOptions->setValidationFunction([](Utils::FancyLineEdit *edit, QString *errorMessage) {
+        const QString value = Utils::globalMacroExpander()->expand(edit->text());
+
+        if (value.isEmpty())
+            return true;
+
+        QJsonParseError parseInfo;
+        const QJsonDocument json = QJsonDocument::fromJson(value.toUtf8(), &parseInfo);
+
+        if (json.isNull()) {
+            if (errorMessage)
+                *errorMessage = tr("Failed to parse JSON at %1: %2")
+                    .arg(parseInfo.offset)
+                    .arg(parseInfo.errorString());
+            return false;
+        }
+        return true;
+    });
+    m_initializationOptions->setText(settings->m_initializationOptions);
+    m_initializationOptions->setPlaceholderText(tr("Language server-specific JSON to pass via "
+                                                   "\"initializationOptions\" field of \"initialize\" "
+                                                   "request."));
 
-    mainLayout->addWidget(new QLabel(tr("Capabilities:")), ++row, 0, Qt::AlignTop);
-    QVector<Client *> clients = LanguageClientManager::clientForSetting(settings);
-    if (clients.isEmpty()) {
-        mainLayout->addWidget(createInfoLabel());
-    } else { // TODO move the capabilities view into a new widget outside of the settings
-        Client *client = clients.first();
-        if (client->state() == Client::Initialized)
-            mainLayout->addWidget(createCapabilitiesView(QJsonValue(client->capabilities())));
-        else
-            mainLayout->addWidget(createInfoLabel(), row, 1);
-        connect(client, &Client::finished, mainLayout, [mainLayout, row, createInfoLabel]() {
-            delete mainLayout->itemAtPosition(row, 1)->widget();
-            mainLayout->addWidget(createInfoLabel(), row, 1);
-        });
-        connect(client, &Client::initialized, mainLayout,
-                [mainLayout, row](
-                    const LanguageServerProtocol::ServerCapabilities &capabilities) {
-                    delete mainLayout->itemAtPosition(row, 1)->widget();
-                    mainLayout->addWidget(createCapabilitiesView(QJsonValue(capabilities)), row, 1);
-                });
-    }
     setLayout(mainLayout);
 }
 
@@ -650,8 +848,8 @@ QString BaseSettingsWidget::name() const
 
 LanguageFilter BaseSettingsWidget::filter() const
 {
-    return {m_mimeTypes->text().split(filterSeparator, QString::SkipEmptyParts),
-                m_filePattern->text().split(filterSeparator, QString::SkipEmptyParts)};
+    return {m_mimeTypes->text().split(filterSeparator, Qt::SkipEmptyParts),
+                m_filePattern->text().split(filterSeparator, Qt::SkipEmptyParts)};
 }
 
 BaseSettings::StartBehavior BaseSettingsWidget::startupBehavior() const
@@ -659,6 +857,11 @@ BaseSettings::StartBehavior BaseSettingsWidget::startupBehavior() const
     return BaseSettings::StartBehavior(m_startupBehavior->currentIndex());
 }
 
+QString BaseSettingsWidget::initializationOptions() const
+{
+    return m_initializationOptions->text();
+}
+
 class MimeTypeModel : public QStringListModel
 {
 public:
@@ -725,7 +928,6 @@ public:
         connect(filter, &QLineEdit::textChanged, proxy, &QSortFilterProxyModel::setFilterWildcard);
         listView->setModel(proxy);
 
-        setWindowFlags(windowFlags() & ~Qt::WindowContextHelpButtonHint);
         setModal(true);
     }
 
@@ -746,7 +948,7 @@ private:
 
 void BaseSettingsWidget::showAddMimeTypeDialog()
 {
-    MimeTypeDialog dialog(m_mimeTypes->text().split(filterSeparator, QString::SkipEmptyParts),
+    MimeTypeDialog dialog(m_mimeTypes->text().split(filterSeparator, Qt::SkipEmptyParts),
                           Core::ICore::dialogParent());
     if (dialog.exec() == QDialog::Rejected)
         return;
@@ -768,13 +970,13 @@ StdIOSettingsWidget::StdIOSettingsWidget(const StdIOSettings *settings, QWidget
     m_executable->setPath(QDir::toNativeSeparators(settings->m_executable));
     mainLayout->addWidget(m_arguments, baseRows + 1, 1);
 
-    auto chooser = new Core::VariableChooser(this);
+    auto chooser = new Utils::VariableChooser(this);
     chooser->addSupportedWidget(m_arguments);
 }
 
 QString StdIOSettingsWidget::executable() const
 {
-    return m_executable->path();
+    return m_executable->filePath().toString();
 }
 
 QString StdIOSettingsWidget::arguments() const
@@ -788,11 +990,16 @@ bool LanguageFilter::isSupported(const Utils::FilePath &filePath, const QString
         return true;
     if (filePattern.isEmpty() && filePath.isEmpty())
         return mimeTypes.isEmpty();
-    auto regexps = Utils::transform(filePattern, [](const QString &pattern){
-        return QRegExp(pattern, Utils::HostOsInfo::fileNameCaseSensitivity(), QRegExp::Wildcard);
+    const QRegularExpression::PatternOptions options
+            = Utils::HostOsInfo::fileNameCaseSensitivity() == Qt::CaseInsensitive
+            ? QRegularExpression::CaseInsensitiveOption : QRegularExpression::NoPatternOption;
+    auto regexps = Utils::transform(filePattern, [&options](const QString &pattern){
+        return QRegularExpression(QRegularExpression::wildcardToRegularExpression(pattern),
+                                  options);
     });
-    return Utils::anyOf(regexps, [filePath](const QRegExp &reg){
-        return reg.exactMatch(filePath.toString()) || reg.exactMatch(filePath.fileName());
+    return Utils::anyOf(regexps, [filePath](const QRegularExpression &reg){
+        return reg.match(filePath.toString()).hasMatch()
+                || reg.match(filePath.fileName()).hasMatch();
     });
 }
 
@@ -801,4 +1008,14 @@ bool LanguageFilter::isSupported(const Core::IDocument *document) const
     return isSupported(document->filePath(), document->mimeType());
 }
 
+bool LanguageFilter::operator==(const LanguageFilter &other) const
+{
+    return this->filePattern == other.filePattern && this->mimeTypes == other.mimeTypes;
+}
+
+bool LanguageFilter::operator!=(const LanguageFilter &other) const
+{
+    return this->filePattern != other.filePattern || this->mimeTypes != other.mimeTypes;
+}
+
 } // namespace LanguageClient
