@@ -56,6 +56,7 @@
 #include <QTimer>
 #include <QShortcut>
 #include <QApplication>
+#include <QScopedPointer>
 
 enum {
     debug = false
@@ -81,14 +82,14 @@ PropertyEditorView::PropertyEditorView(QWidget *parent) :
 {
     m_qmlDir = PropertyEditorQmlBackend::propertyEditorResourcesPath();
 
-    if (Utils::HostOsInfo().isMacHost())
+    if (Utils::HostOsInfo::isMacHost())
         m_updateShortcut = new QShortcut(QKeySequence(Qt::ALT + Qt::Key_F3), m_stackedWidget);
     else
         m_updateShortcut = new QShortcut(QKeySequence(Qt::CTRL + Qt::Key_F3), m_stackedWidget);
     connect(m_updateShortcut, &QShortcut::activated, this, &PropertyEditorView::reloadQml);
 
     m_stackedWidget->setStyleSheet(Theme::replaceCssColors(
-            QString::fromUtf8(Utils::FileReader::fetchQrc(QStringLiteral(":/qmldesigner/stylesheet.css")))));
+        QString::fromUtf8(Utils::FileReader::fetchQrc(":/qmldesigner/stylesheet.css"))));
     m_stackedWidget->setMinimumWidth(340);
     m_stackedWidget->move(0, 0);
     connect(m_stackedWidget, &PropertyEditorWidget::resized, this, &PropertyEditorView::updateSize);
@@ -163,9 +164,9 @@ void PropertyEditorView::changeValue(const QString &name)
             value->setValue(m_selectedNode.id());
             m_locked = false;
             if (!QmlDesigner::ModelNode::isValidId(newId))
-                Core::AsynchronousMessageBox::warning(tr("Invalid Id"),  tr("%1 is an invalid id.").arg(newId));
+                Core::AsynchronousMessageBox::warning(tr("Invalid ID"),  tr("%1 is an invalid ID.").arg(newId));
             else
-                Core::AsynchronousMessageBox::warning(tr("Invalid Id"),  tr("%1 already exists.").arg(newId));
+                Core::AsynchronousMessageBox::warning(tr("Invalid ID"),  tr("%1 already exists.").arg(newId));
         }
         return;
     }
@@ -212,7 +213,6 @@ void PropertyEditorView::changeValue(const QString &name)
         }
     }
 
-    bool forceReset = false;
     if (name == "state" && castedValue.toString() == "base state")
         castedValue = "";
 
@@ -226,12 +226,21 @@ void PropertyEditorView::changeValue(const QString &name)
     if (!value->value().isValid()) { //reset
         removePropertyFromModel(propertyName);
     } else {
-        if (castedValue.isValid() && !castedValue.isNull()) {
+        // QVector*D(0, 0, 0) detects as null variant though it is valid value
+        if (castedValue.isValid() && (!castedValue.isNull()
+                                      || castedValue.type() == QVariant::Vector2D
+                                      || castedValue.type() == QVariant::Vector3D)) {
             commitVariantValueToModel(propertyName, castedValue);
         }
     }
 }
 
+static bool isTrueFalseLiteral(const QString &expression)
+{
+    return (expression.compare("false", Qt::CaseInsensitive) == 0)
+           || (expression.compare("true", Qt::CaseInsensitive) == 0);
+}
+
 void PropertyEditorView::changeExpression(const QString &propertyName)
 {
     PropertyName name = propertyName.toUtf8();
@@ -249,7 +258,7 @@ void PropertyEditorView::changeExpression(const QString &propertyName)
         PropertyName underscoreName(name);
         underscoreName.replace('.', '_');
 
-        QmlObjectNode qmlObjectNode(m_selectedNode);
+        QScopedPointer<QmlObjectNode> qmlObjectNode {QmlObjectNode::getQmlObjectNodeOfCorrectType(m_selectedNode)};
         PropertyEditorValue *value = m_qmlBackEndForCurrentType->propertyValueForName(QString::fromLatin1(underscoreName));
 
         if (!value) {
@@ -257,33 +266,45 @@ void PropertyEditorView::changeExpression(const QString &propertyName)
             return;
         }
 
-        if (qmlObjectNode.modelNode().metaInfo().isValid() && qmlObjectNode.modelNode().metaInfo().hasProperty(name)) {
-            if (qmlObjectNode.modelNode().metaInfo().propertyTypeName(name) == "QColor") {
+        if (qmlObjectNode->modelNode().metaInfo().isValid() && qmlObjectNode->modelNode().metaInfo().hasProperty(name)) {
+            if (qmlObjectNode->modelNode().metaInfo().propertyTypeName(name) == "QColor") {
                 if (QColor(value->expression().remove('"')).isValid()) {
-                    qmlObjectNode.setVariantProperty(name, QColor(value->expression().remove('"')));
+                    qmlObjectNode->setVariantProperty(name, QColor(value->expression().remove('"')));
                     return;
                 }
-            } else if (qmlObjectNode.modelNode().metaInfo().propertyTypeName(name) == "bool") {
-                if (value->expression().compare(QLatin1String("false"), Qt::CaseInsensitive) == 0
-                        || value->expression().compare(QLatin1String("true"), Qt::CaseInsensitive) == 0) {
-                    if (value->expression().compare(QLatin1String("true"), Qt::CaseInsensitive) == 0)
-                        qmlObjectNode.setVariantProperty(name, true);
+            } else if (qmlObjectNode->modelNode().metaInfo().propertyTypeName(name) == "bool") {
+                if (isTrueFalseLiteral(value->expression())) {
+                    if (value->expression().compare("true", Qt::CaseInsensitive) == 0)
+                        qmlObjectNode->setVariantProperty(name, true);
                     else
-                        qmlObjectNode.setVariantProperty(name, false);
+                        qmlObjectNode->setVariantProperty(name, false);
                     return;
                 }
-            } else if (qmlObjectNode.modelNode().metaInfo().propertyTypeName(name) == "int") {
+            } else if (qmlObjectNode->modelNode().metaInfo().propertyTypeName(name) == "int") {
                 bool ok;
                 int intValue = value->expression().toInt(&ok);
                 if (ok) {
-                    qmlObjectNode.setVariantProperty(name, intValue);
+                    qmlObjectNode->setVariantProperty(name, intValue);
                     return;
                 }
-            } else if (qmlObjectNode.modelNode().metaInfo().propertyTypeName(name) == "qreal") {
+            } else if (qmlObjectNode->modelNode().metaInfo().propertyTypeName(name) == "qreal") {
                 bool ok;
                 qreal realValue = value->expression().toDouble(&ok);
                 if (ok) {
-                    qmlObjectNode.setVariantProperty(name, realValue);
+                    qmlObjectNode->setVariantProperty(name, realValue);
+                    return;
+                }
+            } else if (qmlObjectNode->modelNode().metaInfo().propertyTypeName(name) == "QVariant") {
+                bool ok;
+                qreal realValue = value->expression().toDouble(&ok);
+                if (ok) {
+                    qmlObjectNode->setVariantProperty(name, realValue);
+                    return;
+                } else if (isTrueFalseLiteral(value->expression())) {
+                    if (value->expression().compare("true", Qt::CaseInsensitive) == 0)
+                        qmlObjectNode->setVariantProperty(name, true);
+                    else
+                        qmlObjectNode->setVariantProperty(name, false);
                     return;
                 }
             }
@@ -294,8 +315,8 @@ void PropertyEditorView::changeExpression(const QString &propertyName)
             return;
         }
 
-        if (qmlObjectNode.expression(name) != value->expression() || !qmlObjectNode.propertyAffectedByCurrentState(name))
-            qmlObjectNode.setBindingProperty(name, value->expression());
+        if (qmlObjectNode->expression(name) != value->expression() || !qmlObjectNode->propertyAffectedByCurrentState(name))
+            qmlObjectNode->setBindingProperty(name, value->expression());
 
     }); /* end of transaction */
 }
@@ -321,7 +342,7 @@ void PropertyEditorView::exportPopertyAsAlias(const QString &name)
         PropertyName propertyName = aliasName.toUtf8();
         if (rootModelNode().hasProperty(propertyName)) {
             Core::AsynchronousMessageBox::warning(tr("Cannot Export Property as Alias"),
-                                                  tr("Property %1 does already exist for root item.").arg(aliasName));
+                                                  tr("Property %1 does already exist for root component.").arg(aliasName));
             return;
         }
         rootModelNode().bindingProperty(propertyName).setDynamicTypeNameAndExpression("alias", id + "." + name);
@@ -355,12 +376,9 @@ bool PropertyEditorView::locked() const
     return m_locked;
 }
 
-void PropertyEditorView::nodeCreated(const ModelNode &modelNode)
+void PropertyEditorView::currentTimelineChanged(const ModelNode &)
 {
-    if (!m_qmlBackEndForCurrentType->contextObject()->hasActiveTimeline()
-            && QmlTimeline::isValidQmlTimeline(modelNode)) {
-        m_qmlBackEndForCurrentType->contextObject()->setHasActiveTimeline(QmlTimeline::hasActiveTimeline(this));
-    }
+    m_qmlBackEndForCurrentType->contextObject()->setHasActiveTimeline(QmlTimeline::hasActiveTimeline(this));
 }
 
 void PropertyEditorView::updateSize()
@@ -466,12 +484,14 @@ void PropertyEditorView::setupQmlBackend()
         m_stackedWidget->addWidget(currentQmlBackend->widget());
         m_qmlBackendHash.insert(qmlFile.toString(), currentQmlBackend);
 
-        QmlObjectNode qmlObjectNode;
+        QScopedPointer<QmlObjectNode> qmlObjectNode;
         if (m_selectedNode.isValid()) {
-            qmlObjectNode = QmlObjectNode(m_selectedNode);
-            Q_ASSERT(qmlObjectNode.isValid());
+            qmlObjectNode.reset(QmlObjectNode::getQmlObjectNodeOfCorrectType(m_selectedNode));
+            Q_ASSERT(qmlObjectNode->isValid());
+            currentQmlBackend->setup(*qmlObjectNode, currentStateName, qmlSpecificsFile, this);
+        } else {
+            qmlObjectNode.reset(new QmlObjectNode);
         }
-        currentQmlBackend->setup(qmlObjectNode, currentStateName, qmlSpecificsFile, this);
         currentQmlBackend->context()->setContextProperty("finishedNotify", QVariant(false));
         if (specificQmlData.isEmpty())
             currentQmlBackend->contextObject()->setSpecificQmlData(specificQmlData);
@@ -481,14 +501,16 @@ void PropertyEditorView::setupQmlBackend()
         currentQmlBackend->setSource(qmlFile);
         currentQmlBackend->context()->setContextProperty("finishedNotify", QVariant(true));
     } else {
-        QmlObjectNode qmlObjectNode;
+        QScopedPointer<QmlObjectNode> qmlObjectNode;
         if (m_selectedNode.isValid())
-            qmlObjectNode = QmlObjectNode(m_selectedNode);
+            qmlObjectNode.reset(QmlObjectNode::getQmlObjectNodeOfCorrectType(m_selectedNode));
+        else
+            qmlObjectNode.reset(new QmlObjectNode);
 
         currentQmlBackend->context()->setContextProperty("finishedNotify", QVariant(false));
         if (specificQmlData.isEmpty())
             currentQmlBackend->contextObject()->setSpecificQmlData(specificQmlData);
-        currentQmlBackend->setup(qmlObjectNode, currentStateName, qmlSpecificsFile, this);
+        currentQmlBackend->setup(*qmlObjectNode, currentStateName, qmlSpecificsFile, this);
         currentQmlBackend->contextObject()->setGlobalBaseUrl(qmlFile);
         currentQmlBackend->contextObject()->setSpecificQmlData(specificQmlData);
     }
@@ -510,8 +532,10 @@ void PropertyEditorView::commitVariantValueToModel(const PropertyName &propertyN
         RewriterTransaction transaction = beginRewriterTransaction("PropertyEditorView::commitVariantValueToMode");
 
         for (const ModelNode &node : m_selectedNode.view()->selectedModelNodes()) {
-            if (QmlObjectNode::isValidQmlObjectNode(node))
-                QmlObjectNode(node).setVariantProperty(propertyName, value);
+            if (QmlObjectNode::isValidQmlObjectNode(node)) {
+                QScopedPointer<QmlObjectNode>{QmlObjectNode::getQmlObjectNodeOfCorrectType(node)}
+                    ->setVariantProperty(propertyName, value);
+            }
         }
         transaction.commit();
     }
